SystemOrganization addCategory: #Squeak2JS!

TestCase subclass: #S2JParserTest
	instanceVariableNames: 'currentRule'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Squeak2JS'!

!S2JParserTest methodsFor: 'helpers' stamp: 'ag 11/21/2010 11:58'!
assert: aString matches: aRule
	"In most cases, throws an error when failing (hard fail)"
	
	| result |
	result := S2JParser matchString: aString with: aRule.
	self assert: result notNil.
	^ result! !

!S2JParserTest methodsFor: 'helpers' stamp: 'ag 11/21/2010 14:16'!
assert: aString matches: aRule andProduces: anObject
	"In most cases, throws an error when failing (hard fail). #= Checks collections deeply."
	
	| result |
	result := self assert: aString matches: aRule.
	self assert: result = anObject.
	^ result! !

!S2JParserTest methodsFor: 'helpers' stamp: 'ag 11/21/2010 14:18'!
assertFails: aString
	"Checks against the currentRule"
	
	self
		assertFails: aString
		toMatch: self currentRule.! !

!S2JParserTest methodsFor: 'helpers' stamp: 'ag 11/21/2010 14:23'!
assertFails: aString toMatch: aRule
	
	self
		should: [ self assert: aString matches: aRule ]
		raise: OM2Fail
		description: 'Matching should have failed but didn''t!!'.! !

!S2JParserTest methodsFor: 'helpers' stamp: 'ag 11/21/2010 14:15'!
assertMatches: aString
	"Check against the currentRule"
	
	self assert: aString matches: self currentRule.! !

!S2JParserTest methodsFor: 'helpers' stamp: 'ag 11/21/2010 14:15'!
assertMatches: aString andProduces: anObject
	"Check against the currentRule"
	
	self
		assert: aString
		matches: self currentRule
		andProduces: anObject.! !

!S2JParserTest methodsFor: 'accessing' stamp: 'ag 11/21/2010 14:14'!
currentRule
	^ currentRule! !

!S2JParserTest methodsFor: 'accessing' stamp: 'ag 11/21/2010 14:14'!
currentRule: aSymbol
	currentRule := aSymbol! !

!S2JParserTest methodsFor: 'test-tokens' stamp: 'ag 11/21/2010 15:22'!
testBinarySelector
	"Would be cool to find all the binary messages in the image and test them..."
	
	self
		currentRule: #binarySelector;
		
		"special cases: single pipe and single minus"
		assertMatches: '|' andProduces: #|;
		assertMatches: '-' andProduces: #-;
		
		"names starting with a minus"
		assertMatches: '-*' andProduces: #-*;
		assertMatches: '-***' andProduces: #-***;
		assertMatches: '-*%&' andProduces: #-*%&;
		assertMatches: '-*/\' andProduces: #-*/\;
		assertMatches: '-!!?' andProduces: #-!!?;
		assertMatches: '-==' andProduces: #-==;
		
		"mixed characters, check each one"
		assertMatches: '*+' andProduces: #*+;
		assertMatches: '/=\' andProduces: #/=\;
		assertMatches: '~,,,<' andProduces: #~,,,<;
		assertMatches: '>@' andProduces: #>@;
		assertMatches: '%&' andProduces: #%&;
		assertMatches: '?!!' andProduces: #?!!;
		assertMatches: '``' andProduces: #``;
		assertMatches: '+??=' andProduces: #+??=;
		assertMatches: ',!!!!,' andProduces: #,!!!!,;
		
		"Not allowed things"
		assertFails: '';
		assertFails: ' ';
		assertFails: 'w';
		assertFails: '2';
		assertFails: '--';
		assertFails: '||';
		assertFails: '|--';
		assertFails: '%--';
		assertFails: '%||';
		assertFails: '&-&-';
		assertFails: '- ';
		assertFails: ' -';
		assertFails: '& &';
		assertFails: 'w a';
		assertFails: '#a a';
		assertFails: '#aa 2'.! !

!S2JParserTest methodsFor: 'test-tokens' stamp: 'ag 11/21/2010 15:36'!
testCharacter
	
	self
		currentRule: #character;
		
		assertMatches: '$c' andProduces: $c;
		assertMatches: '$&' andProduces: $&;
		assertMatches: '$ ' andProduces: $ ;
		assertMatches: '$`' andProduces: $`;
		
		assertFails: '';
		assertFails: ' ';
		assertFails: 'a';
		assertFails: '`';
		assertFails: ' $a';
		assertFails: '$$$';
		assertFails: '$  ';
		assertFails: '$av';
		assertFails: '$ a'.! !

!S2JParserTest methodsFor: 'test-tokens' stamp: 'ag 11/21/2010 15:36'!
testIdentifier
	
	self
		currentRule: #identifier;
		
		assertMatches: 'abc' andProduces: 'abc';
		assertMatches: 'a1' andProduces: 'a1';
		assertMatches: '1' andProduces: '1';
		assertMatches: 'aaaaaaaa1233333' andProduces: 'aaaaaaaa1233333';
		assertMatches: '' andProduces: '';
		
		assertFails: '';
		assertFails: ' ';
		assertFails: '|';
		assertFails: ':';
		assertFails: '#abc';
		assertFails: '1abc';
		assertFails: '123';
		assertFails: 'abc';
		assertFails: 'abc:';
		assertFails: 'abc<<'.! !

!S2JParserTest methodsFor: 'test-tokens' stamp: 'ag 11/21/2010 15:34'!
testKeywordSelector
	
	self
		currentRule: #keywordSelector;
		
		assertMatches: 'abc:' andProduces: #abc:;
		assertMatches: 'a1:' andProduces: #a1:;
		assertMatches: '1:' andProduces: #1:;
		assertMatches: 'aaaaaaaa1233333:' andProduces: #aaaaaaaa1233333:;
		assertMatches: ':' andProduces: #:;
		
		assertFails: '';
		assertFails: ' ';
		assertFails: '#abc';
		assertFails: '1abc';
		assertFails: '123';
		assertFails: 'abc';
		assertFails: 'abc';
		assertFails: 'abc<<'.! !

!S2JParserTest methodsFor: 'test-tokens' stamp: 'ag 11/21/2010 15:56'!
testNumber
	
	"TODO extend test, when numberS2J gets extended (e.g. -2 0.2 etc.)"
	
	self
		currentRule: #numberS2J;
		
		assertMatches: '1' andProduces: 1;
		assertMatches: '0' andProduces: 0;
		assertMatches: '123' andProduces: 123;
		assertMatches: '112345678' andProduces: 112345678;
		assertMatches: '000' andProduces: 0;
		
		assertFails: '';
		assertFails: ' ';
		assertFails: '#abc';
		assertFails: '1abc';
		assertFails: 'abc';
		assertFails: '<<'.! !

!S2JParserTest methodsFor: 'test-tokens' stamp: 'ag 11/21/2010 16:00'!
testString
	
	"TODO extend test, if stringS2J gets extended to support '' inside of a string"
	
	self
		currentRule: #stringS2J;
		
		assertMatches: '''a''' andProduces: 'a';
		assertMatches: '''''' andProduces: '';
		assertMatches: '''123''' andProduces: '123';
		assertMatches: '''    ''' andProduces: '    ';
		assertMatches: '''"$%&''' andProduces: '"$%&';
		
		assertFails: '';
		assertFails: ' ';
		assertFails: '#abc';
		assertFails: ''' ''a';
		assertFails: ' ''''';
		assertFails: 'a''''';
		assertFails: ''' '' ';
		assertFails: '''as';
		assertFails: 'as''';
		assertFails: ''' '' ''';
		assertFails: '''a '' a''';
		assertFails: '''/'.! !

!S2JParserTest methodsFor: 'test-tokens' stamp: 'ag 11/21/2010 16:05'!
testSymbol
	
	"TODO extend test, if stringS2J gets extended to support '' inside of a string"
	
	self
		currentRule: #symbolS2J;
		
		assertMatches: '#abc' andProduces: #abc;
		assertMatches: '#abc1' andProduces: #abc1;
		assertMatches: '#a' andProduces: #a;
		assertMatches: '#''a''' andProduces: #'a';
		assertMatches: '#''''' andProduces: #'';
		assertMatches: '#''123''' andProduces: #'123';
		assertMatches: '#''    ''' andProduces: #'    ';
		assertMatches: '#''"$%&''' andProduces: #'"$%&';
		
		assertFails: '';
		assertFails: ' ';
		assertFails: 'abc';
		assertFails: '#""';
		assertFails: '#'' ''a';
		assertFails: '# ''''';
		assertFails: '#a''''';
		assertFails: '#'' '' ';
		assertFails: '#''as';
		assertFails: '#as''';
		assertFails: '#'' '' ''';
		assertFails: '#''a '' a''';
		assertFails: '#123';
		assertFails: '###';
		assertFails: '#''/'.! !

!S2JParserTest methodsFor: 'test-tokens' stamp: 'ag 11/21/2010 16:06'!
testUnarySelector
	"Would be cool to find all the unary messages in the image and test them..."
	
	self
		currentRule: #unarySelector;
		
		assertMatches: 'abc' andProduces: #abc;
		assertMatches: 'a1' andProduces: #a1;
		assertMatches: '1' andProduces: #1;
		assertMatches: 'aaaaaaaa1233333' andProduces: #aaaaaaaa1233333;
		assertMatches: '' andProduces: #;
		
		assertFails: '';
		assertFails: ' ';
		assertFails: '|';
		assertFails: ':';
		assertFails: '#abc';
		assertFails: '1abc';
		assertFails: '123';
		assertFails: 'abc';
		assertFails: 'abc:';
		assertFails: 'abc<<'.! !

OMeta2 subclass: #S2JParser
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Squeak2JS'!

!S2JParser class methodsFor: 'matching' stamp: 'ag 11/18/2010 12:16'!
matchString: aString with: aRule
	
	| matcher result |
	matcher := self matcherOn: aString readStream.
	result := matcher
		apply: #empty;
		apply: aRule.
	matcher apply: #end.
	^ result! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/21/2010 13:47'!
binaryExpression =
	
	(primary | unaryExpression | binaryExpression):receiver spaces binarySelector:selector spaces (primary | unaryExpression):argument
	-> [ self node: 'binaryExpression' with: { receiver. selector. argument} ]! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/21/2010 13:46'!
binarySelector =
	
	<$| | ((specialCharacter | $-) (specialCharacter* | empty))>:selector
	-> [ selector asSymbol ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/21/2010 15:47'!
block =
	
	$[
		// arguments
		(((spaces $: identifier:argumentName -> [ argumentName ])+:arguments spaces $| ) | empty)
		// block contents
		spaces
		(temporaries:temporaries | empty)
		spaces
		(statements:statements | empty)
	char($])
	
	-> [ self node: 'block' withoutNils: {
			S2JNode named: 'arguments' withoutNilsNotEmpty: arguments.
			temporaries.
			statements} ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/21/2010 15:47'!
braceExpression =
	
	${ spaces
	((expression:firstExpression
		((spaces $. spaces expression:expr -> [ expr ])*:expressions | empty))
	| empty)
	spaces $}
	-> [ self braceExpressionWith: firstExpression and: expressions ]! !

!S2JParser methodsFor: 'helpers' stamp: 'ag 11/21/2010 13:57'!
braceExpressionWith: firstExpression and: expressions
	
	^ self node: 'braceExpression' withoutNils: 
		((expressions
			ifNotNil: [ expressions ]
			ifNil: [ {} ])
				addFirst: firstExpression;
				yourself)! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/21/2010 15:50'!
character =
	// The method in the super class is #char, don't confuse it!!
	
	$$ <anything>:value
	-> [ value asCharacter ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/21/2010 14:04'!
expression =
	
	// TODO not fixed
	
	// Variable assignments
	(((identifier:identifier (token(':=') | token('_'))) -> [ identifier ])*:variableNames | empty)
	// Actual part of expression
	(messageExpression | primary ):expressionBody
	
	// TODO cascaded-message-expression!!
	// | cascadedMessageExpression
	
	// Create an extra node for the assignments, if any
	-> [ self node: 'expression' withoutNils: { 
			(self node: 'variableAssignments' withoutNilsNotEmpty: variableNames).
			expressionBody} ]! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/21/2010 15:39'!
identifier =
	// Does not read prepended spaces!!
	
	<letter letterOrDigit*>! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/21/2010 13:15'!
keywordSelector =
	
	<identifier $:>:selector
	-> [ selector asSymbol ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/18/2010 15:16'!
messageExpression =
	
	unaryExpression
|	binaryExpression
	
	// TODO keyword-Expressions!!
	//	|	keywordExpression
! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/18/2010 13:10'!
messagePattern =
	
	// Unary
	identifier:unarySelector -> [ S2JNode named: 'unary' with: { unarySelector } ]
	
	// Binary
|	binarySelector:selector identifier:argumentName -> [ S2JNode named: 'binary' with: { selector. argumentName } ]
	
	// Keyword
	// TODO
! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/18/2010 13:12'!
method =
	
	messagePattern:a1 (temporaries:a2 | empty) (primitiveDeclaration:a3 | empty) (statements:a4 | empty)
	-> [ S2JNode named: 'method' withoutNils: { a1. a2. a3. a4 } ]! !

!S2JParser methodsFor: 'helpers' stamp: 'ag 11/18/2010 23:33'!
node: aString with: anOrderedCollection
	
	^ S2JNode named: aString with: anOrderedCollection! !

!S2JParser methodsFor: 'helpers' stamp: 'ag 11/18/2010 23:33'!
node: aString withoutNils: anOrderedCollection
	
	^ S2JNode named: aString withoutNils: anOrderedCollection! !

!S2JParser methodsFor: 'helpers' stamp: 'ag 11/18/2010 23:34'!
node: aString withoutNilsNotEmpty: anOrderedCollection
	
	^ S2JNode named: aString withoutNilsNotEmpty: anOrderedCollection! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/21/2010 15:52'!
numberS2J =
	// Called that way because of super-class-method
	// TODO numbers can be expressed in several different ways!!
	
	<(anything:n ?[n isDigit])+>:result
	-> [ result asNumber ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/21/2010 15:52'!
primary =
	// something, that evaluates to an object
	
	// TODO constant-array-literal; will need to outsource symbol-, number- and character-literals
	
	identifier // variableName / argumentName
|	numberS2J
|	symbolToken:symbolToken -> [ S2JNode named: 'symbol' with: { symbolToken } ]
|	character:character -> [ S2JNode named: 'character' with: { character }]
|	block
|	braceExpression
|	token('(') expression:expr token(')') -> [ expr ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/18/2010 13:15'!
primitiveDeclaration =
	
	// TODO primitive can also take the form <primitive: methodNameString module: moduleNameString>
	
	token('<') token('primitive:') digits:primitiveNumber token('>')
	-> [ S2JNode named: 'primitive' with: { primitiveNumber } ]! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/21/2010 13:46'!
specialCharacter =
	
	$+ | $* | $/ | $\ | $~ | $< | $> | $= | $@ | $% | $& | $? | $!! | $` | $,! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/19/2010 00:54'!
statements =
	
	// Some expressions
	(expression:expr token('.') -> [ expr ] )*:expressions
	
	// The last expression (return statement, not necessary, must not be followed by another expression)
	(((token('^'):returnSign | empty) expression:lastExpression (token('.') | empty) ~expression) | empty)
	
	// Helper method handles the resulting objects
	-> [ self statementsWithExpressions: expressions lastExpression: lastExpression returnSign: returnSign ]! !

!S2JParser methodsFor: 'helpers' stamp: 'ag 11/19/2010 00:57'!
statementsWithExpressions: expressions lastExpression: lastExpression returnSign: returnSign
	"Build the contents of a statements-node from the information gathered while parsing 'statements'"
	
	"Rename the last expression, if it was a return-expression"
	lastExpression notNil & returnSign notNil ifTrue: [ lastExpression name: 'returnExpression' ].
	expressions add: lastExpression.
	^ self node: 'statements' withoutNils: expressions! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/21/2010 15:53'!
stringS2J =
	// Caled that way because of super-class-method
	
	$' <(~$' anything)*>:value $'
	-> [ value ]! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/21/2010 15:53'!
symbolS2J =
	// Called that way because of super-class-method
	
	$# (identifier | stringS2J):value
	-> [ value asSymbol ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/18/2010 13:13'!
temporaries =
	
	token('|') identifier*:variableNames token('|')
	-> [ S2JNode named: 'temporaries' with: variableNames ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/18/2010 17:56'!
unaryExpression =
	
	(primary | unaryExpression):receiver identifier:message
	-> [ S2JNode named: 'unaryExpression' with: {receiver. message} ]! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/21/2010 13:13'!
unarySelector =
	
	identifier:selector
	-> [ selector asSymbol ]! !

OMeta2 subclass: #S2JTranslator
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Squeak2JS'!

!S2JTranslator methodsFor: 'helpers' stamp: 'ag 11/21/2010 15:21'!
node :aString =
	
	// This does not create an S2JNode, but reads one from the input stream and returns it!!
	// For this purpose, this is an ometa-rule with one argument, call like this: node('binarySelector')
	// Other things, that can be found in the result of S2JParser are symbols, strings, numbers etc. (already implemented in superclass)
	// Collections, that are not S2JNodes, should not occur...!!
	
	anything:node ?[node class = S2JNode and: [node name = aString]]
	-> [ node ]! !

OrderedCollection subclass: #S2JNode
	instanceVariableNames: 'name'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Squeak2JS'!

!S2JNode commentStamp: 'ag 11/18/2010 13:09' prior: 0!
This is only a named collection, that is supposed to hold further collections or strings to form a tree!

!S2JNode class methodsFor: 'instance creation' stamp: 'ag 11/18/2010 13:08'!
named: aString with: anOrderedCollection
	
	^ self new
		name: aString;
		addAll: anOrderedCollection;
		yourself! !

!S2JNode class methodsFor: 'instance creation' stamp: 'ag 11/18/2010 13:58'!
named: aString withoutNils: anOrderedCollection
	
	anOrderedCollection ifNil: [ ^ nil ].
	^ self
		named: aString
		with: (anOrderedCollection reject: [ :each | each isNil])! !

!S2JNode class methodsFor: 'instance creation' stamp: 'ag 11/18/2010 13:57'!
named: aString withoutNilsNotEmpty: anOrderedCollection
	"Returns nil, if resulting node would be empty"
	
	| result |
	result := self named: aString withoutNils: anOrderedCollection.
	^ (result isNil | result isEmpty)
		ifTrue: [ nil ]
		ifFalse: [ result ]! !

!S2JNode methodsFor: 'accessors' stamp: 'ag 11/21/2010 12:09'!
= anObject
	
	| result |
	result := super = anObject.
	result := result & (name = anObject name).
	^ result! !

!S2JNode methodsFor: 'accessors' stamp: 'ag 11/18/2010 13:07'!
name
	
	^ name! !

!S2JNode methodsFor: 'accessors' stamp: 'ag 11/18/2010 13:07'!
name: aString
	
	name := aString! !

!S2JNode methodsFor: 'accessors' stamp: 'ag 11/19/2010 00:39'!
printString
	
	| result first |
	first := true.
	result := name, ' ('.
	self do: [ :each |
		first ifTrue: [first := false] ifFalse: [result := result, ', '].
		result := result,  each printString].
	^ result, ')'! !
